@@ -126,7 +126,20 @@ class CatLazyTensor(LazyTensor):
             )
 
     def _expand_batch(self, batch_shape):
-        lazy_tensors = [lazy_tensor._expand_batch(batch_shape) for lazy_tensor in self.lazy_tensors]
+        batch_dim = self.cat_dim + 2
+        if batch_dim < 0:
+            if batch_shape[batch_dim] != self.batch_shape[batch_dim]:
+                raise RuntimeError(
+                    f"Trying to expand a CatLazyTensor in dimension {self.cat_dim}, but this is the concatenated "
+                    f"dimension.\nCurrent shape: {self.shape} - expanded shape: {batch_shape + self.matrix_shape}."
+                )
+            lazy_tensors = []
+            for lazy_tensor in self.lazy_tensors:
+                sub_batch_shape = list(batch_shape).copy()
+                sub_batch_shape[batch_dim] = lazy_tensor.shape[self.cat_dim]
+                lazy_tensors.append(lazy_tensor._expand_batch(sub_batch_shape))
+        else:
+            lazy_tensors = [lazy_tensor._expand_batch(batch_shape) for lazy_tensor in self.lazy_tensors]
         res = self.__class__(*lazy_tensors, dim=self.cat_dim, output_device=self.output_device)
         return res
 
@@ -146,7 +159,7 @@ class CatLazyTensor(LazyTensor):
         lazy_tensors = [self.lazy_tensors[idx] for idx in lazy_tensor_indices]
 
         # Get the new set of indices for each of the LazyTensors
-        switch_tensor = does_switch_tensor.nonzero().squeeze(-1)
+        switch_tensor = does_switch_tensor.nonzero(as_tuple=False).squeeze(-1)
         split_sizes = (switch_tensor[1:] - switch_tensor[:-1]).tolist()
         sub_indices = zip(
             *[
@@ -199,7 +212,7 @@ class CatLazyTensor(LazyTensor):
             lazy_tensors = [self.lazy_tensors[idx] for idx in lazy_tensor_indices]
 
             # Get the new set of indices for each of the LazyTensors
-            switch_tensor = does_switch_tensor.nonzero().squeeze(-1)
+            switch_tensor = does_switch_tensor.nonzero(as_tuple=False).squeeze(-1)
             split_sizes = (switch_tensor[1:] - switch_tensor[:-1]).tolist()
             sub_indices = zip(
                 *[
