@@ -278,6 +278,7 @@ class LazyTensorTestCase(RectangularLazyTensorTestCase):
     skip_slq_tests = False
     should_call_cg = True
     should_call_lanczos = True
+    should_call_lanczos_diagonalization = True
     tolerances = {
         **RectangularLazyTensorTestCase.tolerances,
         "cholesky": {"rtol": 1e-3, "atol": 1e-5},
@@ -337,10 +338,11 @@ class LazyTensorTestCase(RectangularLazyTensorTestCase):
             else:
                 self.assertFalse(linear_cg_mock.called)
 
-    def _test_inv_quad_logdet(self, reduce_inv_quad=True, cholesky=False):
+    def _test_inv_quad_logdet(self, reduce_inv_quad=True, cholesky=False, lazy_tensor=None):
         if not self.__class__.skip_slq_tests:
             # Forward
-            lazy_tensor = self.create_lazy_tensor()
+            if lazy_tensor is None:
+                lazy_tensor = self.create_lazy_tensor()
             evaluated = self.evaluate_lazy_tensor(lazy_tensor)
             flattened_evaluated = evaluated.view(-1, *lazy_tensor.matrix_shape)
 
@@ -404,6 +406,61 @@ class LazyTensorTestCase(RectangularLazyTensorTestCase):
                 actual[..., i, i] = actual[..., i, i] + other_diag[..., i]
             self.assertAllClose(res, actual, **self.tolerances["diag"])
 
+    def test_add_low_rank(self):
+        lazy_tensor = self.create_lazy_tensor()
+        lazy_tensor = self.create_lazy_tensor()
+        evaluated = self.evaluate_lazy_tensor(lazy_tensor)
+        new_rows = torch.randn(*lazy_tensor.shape[:-1], 3)
+
+        summed_lt = evaluated + new_rows.matmul(new_rows.transpose(-1, -2))
+        new_lt = lazy_tensor.add_low_rank(new_rows)
+
+        # check that the concatenation is okay
+        self.assertAllClose(new_lt.evaluate(), summed_lt)
+
+        # check that the root approximation is close
+        rhs = torch.randn(lazy_tensor.size(-1))
+        summed_rhs = summed_lt.matmul(rhs)
+        root_rhs = new_lt.root_decomposition().matmul(rhs)
+        self.assertAllClose(root_rhs, summed_rhs, **self.tolerances["root_decomposition"])
+
+        # check that the inverse root decomposition is close
+        summed_solve = torch.solve(rhs.unsqueeze(-1), summed_lt)[0].squeeze(-1)
+        root_inv_solve = new_lt.root_inv_decomposition().matmul(rhs)
+        self.assertAllClose(root_inv_solve, summed_solve, **self.tolerances["root_inv_decomposition"])
+
+    def test_cat_rows(self):
+        lazy_tensor = self.create_lazy_tensor()
+        evaluated = self.evaluate_lazy_tensor(lazy_tensor)
+
+        for batch_shape in (torch.Size(), torch.Size([2])):
+            new_rows = 1e-4 * torch.randn(*batch_shape, *lazy_tensor.shape[:-2], 1, lazy_tensor.shape[-1])
+            new_point = torch.rand(*batch_shape, *lazy_tensor.shape[:-2], 1, 1)
+
+            # we need to expand here to be able to concat (this happens automatically in cat_rows)
+            cat_col1 = torch.cat((evaluated.expand(*batch_shape, *evaluated.shape), new_rows), dim=-2)
+            cat_col2 = torch.cat((new_rows.transpose(-1, -2), new_point), dim=-2)
+
+            concatenated_lt = torch.cat((cat_col1, cat_col2), dim=-1)
+            new_lt = lazy_tensor.cat_rows(new_rows, new_point)
+
+            # check that the concatenation is okay
+            self.assertAllClose(new_lt.evaluate(), concatenated_lt)
+
+            # check that the root approximation is close
+            rhs = torch.randn(lazy_tensor.size(-1) + 1)
+            concat_rhs = concatenated_lt.matmul(rhs)
+            root_rhs = new_lt.root_decomposition().matmul(rhs)
+            self.assertAllClose(root_rhs, concat_rhs, **self.tolerances["root_decomposition"])
+
+            # check that the inverse root decomposition is close
+            concat_solve = torch.solve(rhs.unsqueeze(-1), concatenated_lt).solution.squeeze(-1)
+            root_inv_solve = new_lt.root_inv_decomposition().matmul(rhs)
+            self.assertLess(
+                (root_inv_solve - concat_solve).norm() / concat_solve.norm(),
+                self.tolerances["root_inv_decomposition"]["rtol"],
+            )
+
     def test_cholesky(self):
         lazy_tensor = self.create_lazy_tensor()
         evaluated = self.evaluate_lazy_tensor(lazy_tensor)
@@ -520,8 +577,52 @@ class LazyTensorTestCase(RectangularLazyTensorTestCase):
             else:
                 self.assertFalse(lanczos_mock.called)
 
+    def test_diagonalization(self, symeig=False):
+        _wrapped_lanczos = MagicMock(wraps=gpytorch.utils.lanczos.lanczos_tridiag)
+        with patch("gpytorch.utils.lanczos.lanczos_tridiag", new=_wrapped_lanczos) as lanczos_mock:
+            lazy_tensor = self.create_lazy_tensor()
+            test_mat = torch.randn(*lazy_tensor.batch_shape, lazy_tensor.size(-1), 5)
+            with gpytorch.settings.max_cholesky_size(math.inf if symeig else 0):
+                evals, evecs = lazy_tensor.diagonalization()
+                evecs = evecs.evaluate()
+                approx = evecs.matmul(torch.diag_embed(evals)).matmul(evecs.transpose(-2, -1))
+                res = approx.matmul(test_mat)
+                actual = lazy_tensor.matmul(test_mat)
+                self.assertAllClose(res, actual, rtol=0.05)
+
+            # Make sure that we're calling the correct function
+            if not symeig and self.__class__.should_call_lanczos_diagonalization:
+                self.assertTrue(lanczos_mock.called)
+            else:
+                self.assertFalse(lanczos_mock.called)
+
+    def test_diagonalization_symeig(self):
+        return self.test_diagonalization(symeig=True)
+
+    def _test_triangular_lazy_tensor_inv_quad_logdet(self):
+        # now we need to test that a second cholesky isn't being called in the inv_quad_logdet
+        with gpytorch.settings.max_cholesky_size(math.inf):
+            lazy_tensor = self.create_lazy_tensor()
+            rootdecomp = lazy_tensor.root_decomposition()
+
+            if isinstance(rootdecomp, gpytorch.lazy.CholLazyTensor):
+                chol = lazy_tensor.root_decomposition().root.clone()
+                gpytorch.utils.memoize.clear_cache_hook(lazy_tensor)
+                gpytorch.utils.memoize.add_to_cache(
+                    lazy_tensor, "root_decomposition", gpytorch.lazy.RootLazyTensor(chol)
+                )
+
+                _wrapped_cholesky = MagicMock(wraps=torch.cholesky)
+                with patch("torch.cholesky", new=_wrapped_cholesky) as cholesky_mock:
+                    self._test_inv_quad_logdet(reduce_inv_quad=True, cholesky=True, lazy_tensor=lazy_tensor)
+                self.assertFalse(cholesky_mock.called)
+
     def test_root_decomposition_cholesky(self):
-        return self.test_root_decomposition(cholesky=True)
+        # first test if the root decomposition is accurate
+        self.test_root_decomposition(cholesky=True)
+
+        # now test that a second cholesky isn't being called in the inv_quad_logdet
+        self._test_inv_quad_logdet()
 
     def test_root_inv_decomposition(self):
         lazy_tensor = self.create_lazy_tensor()
